{ File   : CompSLZX.dpr
  Author : Deniz Oezmen
  Created: 2010-10-16, based on MakSPR.dpr created 2010-10-06
  Changed: 2010-10-17

  SLZX compression functions.
}
unit CompSLZX;

interface

uses
  Shared, MemSearch;

const
  CompMethod_SLZX: array[0..3] of AnsiChar = 'SLZX';

function Compress_SLZX(const InBuffer: TBuffer;
  out OutBuffer: TBuffer; var CompSuccess: Boolean): Longint;

implementation

type
  TSLZXHeader = packed record
    ID        : array[0..3] of AnsiChar;
    UncompSize: Longword;
  end;

function Compress_SLZX(const InBuffer: TBuffer;
  out OutBuffer: TBuffer; var CompSuccess: Boolean): Longint;
var
  SLZXHeader: TSLZXHeader;
  InIdx,
  OutIdx,
  HisIdx,
  FlagIdx,
  HisSize,
  PrefixSize,
  MatchPos,
  MatchLen,
  NumOps    : Longint;
  Flag      : Byte;
  OfsLen    : Word;
begin
  CompSuccess := False;
  Result := SizeOf(SLZXHeader);

  Move(CompMethod_SLZX, SLZXHeader.ID, SizeOf(SLZXHeader.ID));
  SLZXHeader.UncompSize := Length(InBuffer);

  SetLength(OutBuffer, SizeOf(SLZXHeader) + SizeOf(SLZXHeader.UncompSize));
  Move(SLZXHeader, OutBuffer[0], SizeOf(SLZXHeader));
  OutIdx := SizeOf(SLZXHeader);
  Move(SLZXHeader.UncompSize, OutBuffer[OutIdx], SizeOf(SLZXHeader.UncompSize));
  Inc(OutIdx, SizeOf(SLZXHeader.UncompSize));

  NumOps := 0;
  Flag := 0;
  PrefixSize := 31;

  InIdx := 0;
  HisIdx := 0;
  FlagIdx := 0;
  while InIdx <= High(InBuffer) do
  begin
    if NumOps = 0 then
    begin
      // skip to reserve space for the flag byte
      Flag := 0;
      FlagIdx := OutIdx;
      SetLength(OutBuffer, Length(OutBuffer) + 1);
      Inc(OutIdx)
    end;

    // start scrolling the history window as soon as enough data has been seen
    HisSize := InIdx - HisIdx;
    if HisSize > 2048 then
    begin
      HisIdx := InIdx - 2048;
      HisSize := 2048
    end;

    // shorten the prefix when we reach the end of the input buffer
    if High(InBuffer) - InIdx + 1 < 31 then
      PrefixSize := High(InBuffer) - InIdx + 1;

    // search for a maximum-length prefix match of the upcoming data in the
    // history window
    SearchMaxPrefix(SearchMethod_KnuthMorrisPratt, @InBuffer[InIdx],
      @InBuffer[HisIdx], PrefixSize, HisSize, MatchPos, MatchLen);

    // FIXME:
    // 1. Why are the original SLZX data blocks compressed worse than using this
    //    relatively simple implementation?
    // 2. Does the game's decompression engine support the special case of
    //    crossing pointers for a sequence of singletons?

    // a significant match has been found
    if (MatchLen > 3) or
      ((MatchLen = 3) and (InIdx + MatchLen <= High(InBuffer))) then
    begin
      CompSuccess := True;

      if InIdx + MatchLen > High(InBuffer) then
        Dec(MatchLen, 1);

      // output history reference
      OfsLen := MatchPos or (MatchLen shl 11);

      SetLength(OutBuffer, Length(OutBuffer) + SizeOf(OfsLen));
      Move(OfsLen, OutBuffer[OutIdx], SizeOf(OfsLen));
      Inc(InIdx, MatchLen);
      Inc(OutIdx, SizeOf(OfsLen));

      // emit an additional character
      SetLength(OutBuffer, Length(OutBuffer) + 1);
      Move(InBuffer[InIdx], OutBuffer[OutIdx], 1);
      Inc(InIdx, 1);
      Inc(OutIdx, 1);

      Flag := (Flag shr 1) or $80
    end
    else
    begin
      // no match, output a single character
      SetLength(OutBuffer, Length(OutBuffer) + 1);
      Move(InBuffer[InIdx], OutBuffer[OutIdx], 1);
      Inc(InIdx, 1);
      Inc(OutIdx, 1);

      Flag := Flag shr 1
    end;

    Inc(NumOps);

    if (NumOps = 8) or (InIdx > High(InBuffer)) then
    begin
      Flag := Flag shr (8 - NumOps);

      if NumOps < 8 then
      begin
        // fill rest of buffer with pseudo data
        OutIdx := High(OutBuffer);
        SetLength(OutBuffer, Length(OutBuffer) + 8 - NumOps);
        FillChar(OutBuffer[OutIdx + 1], 8 - NumOps, 0)
      end;

      // output final flag byte
      OutBuffer[FlagIdx] := Flag;
      NumOps := 0
    end
  end;
end;

end.
