{ File   : CompSSHC.dpr
  Author : Deniz Oezmen
  Created: 2010-10-16, based on MakSPR.dpr created 2010-10-06
  Changed: 2010-10-16

  SSHC compression functions.
}
unit CompSSHC;

interface

uses
  Shared;

const
  CompMethod_SSHC: array[0..3] of AnsiChar = 'SSHC';

function Compress_SSHC(const InBuffer: TBuffer;
  out OutBuffer: TBuffer): Longint;

implementation

uses
  Heap2;

const
  Null : Word = $0fff;
  Start: Word = $fff6;

type
  THistogramRec = record
    Parent    : Word;
    LeftChild : Word;
    RightChild: Word;
    Value     : Word;
    Index     : Longint;
    LeftIdx   : Longint;
    RightIdx  : Longint;
    Num       : Longint;
    CodeLen   : Longint;
    Code      : TBuffer;
  end;
  THistogram = array of THistogramRec;

  TSSHCHeader = packed record
    ID        : array[0..3] of Char;
    UncompSize: Longword;
    CompSize  : Longword;
    NumNodes  : Longword;
    NumBits   : Longword;
  end;

  THuffNode = packed record
    Parent    : Word;
    LeftChild : Word;
    RightChild: Word;
    Value     : Word;
  end;
  THuffTree = packed array of THuffNode;

var
  Histogram: THistogram;  

function CompareHistogram(const PayloadA,
  PayloadB: THeapPayload): THeapCompare;
var
  HistogramRecA,
  HistogramRecB: THistogramRec;
begin
  HistogramRecA := Histogram[PInteger(PayloadA)^];
  HistogramRecB := Histogram[PInteger(PayloadB)^];

  // switch comparison results to create a minimum-heap
  if HistogramRecA.Num = HistogramRecB.Num then
  begin
    if HistogramRecA.Value = HistogramRecB.Value then
      Result := Compare_Eq
    else if HistogramRecA.Value < HistogramRecB.Value then
      Result := Compare_Gt
    else
      Result := Compare_Lt
  end
  else if HistogramRecA.Num < HistogramRecB.Num then
    Result := Compare_Gt
  else
    Result := Compare_Lt
end;

procedure DeleteHistogram(var Payload: THeapPayload);
begin
  Dispose(PInteger(Payload));
  Payload := nil
end;

procedure CopyHistogram(const Source: THeapPayload; var Dest: THeapPayload);
begin
  New(PInteger(Dest));
  PInteger(Dest)^ := PInteger(Source)^
end;

procedure CopyToHuffTree(const HistRec: THistogramRec;
  var HuffTree: THuffTree);
begin
  SetLength(HuffTree, Length(HuffTree) + 1);
  with HuffTree[High(HuffTree)] do
  begin
    Parent     := HistRec.Parent;
    LeftChild  := HistRec.LeftChild;
    RightChild := HistRec.RightChild;
    Value      := HistRec.Value
  end
end;

procedure Traverse(const Histogram: THistogram; HistIdx: Longint);
var
  ChildIdx,
  CodeIdx : Longint;
begin
  if HistIdx = -1 then
    Exit;

  ChildIdx := Histogram[HistIdx].LeftIdx;
  if ChildIdx <> -1 then
  begin
    Histogram[ChildIdx].CodeLen := Histogram[HistIdx].CodeLen;
    Histogram[ChildIdx].Code := Copy(Histogram[HistIdx].Code);

    if (Histogram[ChildIdx].CodeLen and $07) = 0 then
    begin
      SetLength(Histogram[ChildIdx].Code,
        Length(Histogram[ChildIdx].Code) + 1);
      Histogram[ChildIdx].Code[High(Histogram[ChildIdx].Code)] := 0
    end;

    Inc(Histogram[ChildIdx].CodeLen);

    Traverse(Histogram, ChildIdx)
  end;

  ChildIdx := Histogram[HistIdx].RightIdx;
  if ChildIdx <> -1 then
  begin
    Histogram[ChildIdx].CodeLen := Histogram[HistIdx].CodeLen;
    Histogram[ChildIdx].Code := Copy(Histogram[HistIdx].Code);

    if (Histogram[ChildIdx].CodeLen and $07) = 0 then
    begin
      SetLength(Histogram[ChildIdx].Code,
        Length(Histogram[ChildIdx].Code) + 1);
      Histogram[ChildIdx].Code[High(Histogram[ChildIdx].Code)] := 0
    end;

    CodeIdx := High(Histogram[ChildIdx].Code);

    Histogram[ChildIdx].Code[CodeIdx] := Histogram[ChildIdx].Code[CodeIdx] or
      (1 shl (Histogram[ChildIdx].CodeLen and $07));
    Inc(Histogram[ChildIdx].CodeLen);

    Traverse(Histogram, ChildIdx)
  end
end;

procedure AddCode(var Buffer: TBuffer; var NumBits: Longint;
  const Code: TBuffer; CodeLen: Longint);
var
  i,
  BufIdx,
  NewBits: Longint;
begin
  BufIdx := High(Buffer);

  NewBits := CodeLen - (8 - (NumBits and $07)) and $07;

  if NewBits > 0 then
  begin
    SetLength(Buffer,
      Length(Buffer) + (NewBits shr 3) + Byte(NewBits and $07 > 0));
    FillChar(Buffer[BufIdx + 1], High(Buffer) - BufIdx, 0)
  end;

  for i := 0 to CodeLen - 1 do
  begin
    if NumBits and $07 = 0 then
      Inc(BufIdx);

    Buffer[BufIdx] := Buffer[BufIdx] or
      ((Code[i shr 3] shr (i and $07)) and 1) shl (NumBits and $07);
    Inc(NumBits)
  end
end;

function Compress_SSHC(const InBuffer: TBuffer;
  out OutBuffer: TBuffer): Longint;
var
  i,
  InIdx,
  HistIdx,
  CompBits  : Longint;
  HistIdx1,
  HistIdx2  : PInteger;
  NodeVal   : Word;
  Heap      : THeap;
  HuffTree  : THuffTree;
  SSHCHeader: TSSHCHeader;
  QuickAcc  : array[0..255] of Longint;
begin
  Result := 0;
  SetLength(OutBuffer, 0);

  if Length(InBuffer) = 0 then
    Exit;

  // create histogram of input data
  SetLength(Histogram, 0);
  for InIdx := 0 to High(InBuffer) do
  begin
    HistIdx := -1;
    for i := 0 to High(Histogram) do
      if Histogram[i].Value = InBuffer[InIdx] then
        HistIdx := i;

    if HistIdx < 0 then
    begin
      SetLength(Histogram, Length(Histogram) + 1);
      HistIdx := High(Histogram);
      with Histogram[HistIdx] do
      begin
        Parent     := Null;
        LeftChild  := Null;
        RightChild := Null;
        Value      := InBuffer[InIdx];
        Index      := HistIdx;
        LeftIdx    := -1;
        RightIdx   := -1;
        Num        := 0;
        CodeLen    := 0;
        SetLength(Code, 0)
      end;

      QuickAcc[Histogram[HistIdx].Value] := HistIdx
    end;

    Inc(Histogram[HistIdx].Num, 1)
  end;

  // insert all leaf nodes into a heap
  Heap := THeap.Create(CompareHistogram, DeleteHistogram, CopyHistogram);
  for i := 0 to High(Histogram) do
    Heap.Insert(@i);

  // construct the inner tree nodes
  NodeVal := Start;
  SetLength(HuffTree, 0);
  while True do
  begin
    // get first minimal node
    Heap.GetNext(Pointer(HistIdx1));
    Heap.Delete;

    // get second minimal node: if this does not exist, we have reached the
    // root node
    if not Heap.GetNext(Pointer(HistIdx2)) then
    begin
      Heap.Delete;
      Break
    end;
    Heap.Delete;

    // create a new parent for both nodes
    SetLength(Histogram, Length(Histogram) + 1);
    HistIdx := High(Histogram);
    with Histogram[HistIdx] do
    begin
      Parent     := Null;
      LeftChild  := Histogram[HistIdx1^].Value;
      RightChild := Histogram[HistIdx2^].Value;
      Value      := NodeVal;
      Index      := HistIdx;
      LeftIdx    := Histogram[HistIdx1^].Index;
      RightIdx   := Histogram[HistIdx2^].Index;
      Num        := Histogram[HistIdx1^].Num + Histogram[HistIdx2^].Num
    end;

    Heap.Insert(@HistIdx);

    // connect both nodes to their parent
    Histogram[HistIdx1^].Parent := NodeVal;
    Histogram[HistIdx2^].Parent := NodeVal;

    Dec(NodeVal);
    
    // copy the finalized nodes to the game-internal structure
    CopyToHuffTree(Histogram[HistIdx1^], HuffTree);
    CopyToHuffTree(Histogram[HistIdx2^], HuffTree)
  end;

  // copy the root node to the game-internal structure
  CopyToHuffTree(Histogram[HistIdx1^], HuffTree);
  Heap.Free;

  // copy final tree to output
  SetLength(OutBuffer, SizeOf(SSHCHeader) +
    Length(HuffTree) * SizeOf(THuffNode));
  Move(HuffTree[0], OutBuffer[SizeOf(SSHCHeader)],
    Length(HuffTree) * SizeOf(THuffNode));

  // generate the Huffman code sequences
  Traverse(Histogram, High(Histogram));

  // encode the input data
  CompBits := 0;
  for InIdx := 0 to High(InBuffer) do
    AddCode(OutBuffer, CompBits, Histogram[QuickAcc[InBuffer[InIdx]]].Code,
      Histogram[QuickAcc[InBuffer[InIdx]]].CodeLen);

  // finalize header
  with SSHCHeader do
  begin
    Move(CompMethod_SSHC, ID, SizeOf(ID));
    UncompSize := Length(InBuffer);
    CompSize := Length(OutBuffer) - SizeOf(SSHCHeader);
    NumNodes := Length(HuffTree);
    NumBits := CompBits
  end;

  Move(SSHCHeader, OutBuffer[0], SizeOf(SSHCHeader));

  // clean up
  SetLength(HuffTree, 0);

  for i := 0 to High(Histogram) do
    SetLength(Histogram[i].Code, 0);
  SetLength(Histogram, 0);

  Result := SizeOf(SSHCHeader)
end;

end.
