{ File   : Heap2.pas
  Author : Deniz Oezmen
  Created: 2010-10-15, based on Heap.pas created 2010-10-14
  Changed: 2010-10-16

  Generic maximum-heap functions.
}
unit Heap2;

interface

type
  THeapPayload = Pointer;

  THeapCompare = (Compare_Lt, Compare_Eq, Compare_Gt);
  THeapCompareFunc = function(const PayloadA,
    PayloadB: THeapPayload): THeapCompare;
  THeapDeleteFunc = procedure(var Payload: THeapPayload);
  THeapCopyFunc = procedure(const Source: THeapPayload; var Dest: THeapPayload);

  THeap = class(TObject)
  protected
    FHeap: array of THeapPayload;
    FComparePayload: THeapCompareFunc;
    FDeletePayload: THeapDeleteFunc;
    FCopyPayload: THeapCopyFunc;
    procedure UpHeap(Idx: Integer);
    procedure DownHeap(Idx: Integer);
  public
    constructor Create(const ComparePayload: THeapCompareFunc;
      const DeletePayload: THeapDeleteFunc; const CopyPayload: THeapCopyFunc);
    destructor Destroy; override;
    procedure Insert(const Payload: THeapPayload);
    procedure Delete;
    function GetNext(out Payload: THeapPayload): Boolean;
  end;

implementation

// INTERNAL FUNCTIONS //////////////////////////////////////////////////////////

(* heap manipulation routines *************************************************)
procedure THeap.UpHeap(Idx: Integer);
var
  Payload: THeapPayload;
begin
  if (Idx < 1) or (Idx > High(FHeap)) then
    Exit;

  // store the offending node
  FCopyPayload(FHeap[Idx], Payload);

  // while the node is larger than its parent, move it up
  while (Idx shr 1 > 0) and
    (FComparePayload(Payload, FHeap[Idx shr 1]) = Compare_Gt) do
  begin
    FDeletePayload(FHeap[Idx]);
    FCopyPayload(FHeap[Idx shr 1], FHeap[Idx]);
    Idx := Idx shr 1
  end;

  // the node has reached its final position
  FDeletePayload(FHeap[Idx]);
  FCopyPayload(Payload, FHeap[Idx]);
  FDeletePayload(Payload)
end;

procedure THeap.DownHeap(Idx: Integer);
var
  i,
  Max    : Integer;
  Payload: THeapPayload;
begin
  if (Idx < 1) or (Idx > High(FHeap)) then
    Exit;

  // store the offending node
  FCopyPayload(FHeap[Idx], Payload);

  Max := High(FHeap);
  while Idx <= Max shr 1 do
  begin
    // compare node with both of its children
    i := Idx shl 1;
    if i < Max then
      if FComparePayload(FHeap[i], FHeap[i + 1]) = Compare_Lt then
        Inc(i);
    if FComparePayload(FHeap[i], Payload) = Compare_Lt then
      Break;

    // switch node with its larger child
    FDeletePayload(FHeap[Idx]);
    FCopyPayload(FHeap[i], FHeap[Idx]);
    Idx := i
  end;

  // the node has reached its final position
  FDeletePayload(FHeap[Idx]);
  FCopyPayload(Payload, FHeap[Idx]);
  FDeletePayload(Payload)
end;

// EXTERNAL FUNCTIONS //////////////////////////////////////////////////////////

(* constructors and destructors ***********************************************)
constructor THeap.Create(const ComparePayload: THeapCompareFunc;
  const DeletePayload: THeapDeleteFunc; const CopyPayload: THeapCopyFunc);
begin
  inherited Create;

  SetLength(FHeap, 1);
  FComparePayload := ComparePayload;
  FDeletePayload := DeletePayload;
  FCopyPayload := CopyPayload
end;

destructor THeap.Destroy;
var
  i: Integer;
begin
  for i := 1 to High(FHeap) do
    FDeletePayload(FHeap[i]);

  SetLength(FHeap, 0);

  FComparePayload := nil;
  FDeletePayload := nil;
  FCopyPayload := nil;

  inherited Destroy
end;

(* heap service routines ******************************************************)
procedure THeap.Insert(const Payload: THeapPayload);
var
  LastIdx: Integer;
begin
  // insert the element at the very last heap position
  SetLength(FHeap, Length(FHeap) + 1);
  LastIdx := High(FHeap);

  FCopyPayload(Payload, FHeap[LastIdx]);

  // move the inserted element up until it reaches a valid position
  UpHeap(LastIdx)
end;

procedure THeap.Delete;
begin
  if High(FHeap) < 1 then
    Exit;

  // remove the largest (first) element
  FDeletePayload(FHeap[1]);
  if High(FHeap) > 1 then
  begin
    // if something remains, replace it with the last
    FCopyPayload(FHeap[High(FHeap)], FHeap[1]);
    FDeletePayload(FHeap[High(FHeap)])
  end;

  SetLength(FHeap, Length(FHeap) - 1);

  // move the replaced root down until it reaches a valid position
  DownHeap(1)
end;

function THeap.GetNext(out Payload: THeapPayload): Boolean;
begin
  Result := (High(FHeap) >= 1);

  if Result then
    FCopyPayload(FHeap[1], Payload)
end;

end.

